/* poller - front-end for waiting for I/O functions
   Copyright (C) 2010  Marek Behun

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#include "config.h"
#include "poller.h"

#ifdef HAVE_EPOLL_WAIT

int poller_epollfd = -1;
int poller_nfds = 0;

int
poller_start ()
{
  if (poller_epollfd >= 0)
    {
      errno = EEXIST;
      return (-1);
    }
  poller_epollfd = epoll_create (1024);
  return (poller_epollfd >= 0 ? 0 : -1);
}

int
poller_end ()
{
  if (poller_epollfd < 0)
    {
      errno = ENOENT;
      return (-1);
    }

  close (poller_epollfd);
  poller_epollfd = -1;
  return 0;
}

int
poller_add (int fd, pollerev_t events, pollerdata_t data)
{
  struct epoll_event event;

  event.events = events;
  event.data = data;
  if (!epoll_ctl (poller_epollfd, EPOLL_CTL_ADD, fd, &event))
    {
      ++ poller_nfds;
      return 0;
    }
  return (-1);
}

int
poller_modify (int fd, pollerev_t events, pollerdata_t data)
{
  struct epoll_event event;

  event.events = events;
  event.data = data;
  return epoll_ctl (poller_epollfd, EPOLL_CTL_MOD, fd, &event);
}

int
poller_remove (int fd)
{
  /* There is a bug in kernel versions before 2.6.9 in the epoll_ctl
     call: EPOLL_CTL_DEL require the event pointer to be a non-NULL
     pointer, even though it is ignored. That forces us to pass a
     pointer (&poller_epollfd) as the fourth argument.  */

  if (!epoll_ctl (poller_epollfd, EPOLL_CTL_DEL, fd, (struct epoll_event *) &poller_epollfd))
    -- poller_nfds;
  return poller_nfds;
}

int
poller_poll (int timeout, pollercb_t callback)
{
  struct epoll_event evs[8];
  int i, res;

  res = epoll_wait (poller_epollfd, &evs, 8, timeout);
  if (res <= 0)
    return res;

  for (i = 0; i < res; i++)
    callback (evs[i].data, evs[i].events);

  return res;
}

#else /* not HAVE_EPOLL_WAIT */

struct pollfd * poller_fds = NULL;
pollerdata_t * poller_data = NULL;
int poller_nfds = 0;

int
poller_start ()
{
  if (poller_fds != NULL || poller_data != NULL || poller_nfds != 0)
    {
      errno = EEXIST;
      return (-1);
    }
  return 0;
}

int
poller_end ()
{
  if (poller_fds == NULL || poller_data == NULL)
    {
      errno = ENOENT;
      return (-1);
    }
  free (poller_fds);
  free (poller_data);
  poller_fds = NULL;
  poller_data = NULL;
  poller_nfds = 0;
  return 0;
}

int
poller_add (int fd, pollerev_t events, pollerdata_t data)
{
  struct pollfd * new_fds;
  pollerdata_t * new_fds_data;

  if (fd < 0)
    {
      errno = EBADF;
      return (-1);
    }

  /* Now here, according to POSIX 2004 realloc's man page (I did not
     look in older specifications) and also according to my Linux man
     page, we need not to use malloc() explicitly. Quoting, 'If ptr is
     a null pointer, realloc() shall be equivalent to malloc() for the
     specified size.'.  */

  new_fds = realloc (poller_fds, sizeof (struct pollfd) * ++ poller_nfds);
  if (new_fds == NULL)
    {
      -- poller_nfds;
      errno = ENOMEM;
      return (-1);
    }

  new_fds_data = realloc (poller_data, sizeof (pollerdata_t) * poller_nfds);
  if (new_fds_data == NULL)
    {
      -- poller_nfds;
      poller_fds = realloc (new_fds, sizeof (struct pollfd) * poller_nfds);
      errno = ENOMEM;
      return (-1);
    }

  poller_fds = new_fds;
  poller_data = new_fds_data;

  poller_fds[poller_nfds-1].fd = fd;
  poller_fds[poller_nfds-1].events = events;
  poller_data[poller_nfds-1] = data;

  return 0;
}

int
poller_modify (int fd, pollerev_t events, pollerdata_t data)
{
  int i;

  if (fd < 0)
    {
      errno = EBADF;
      return (-1);
    }

  for (i = 0; i < poller_nfds; i++)
    if (poller_fds[i].fd == fd)
      break;

  if (i == poller_nfds)
    {
      errno = ENOENT;
      return (-1);
    }

  poller_fds[i].events = events;
  poller_data[i] = data;

  return 0;
}

int
poller_remove (int fd)
{
  int i;

  if (fd < 0)
    {
      errno = EBADF;
      goto end;
    }

  for (i = 0; i < poller_nfds; i++)
    if (poller_fds[i].fd == fd)
      break;

  if (i == poller_nfds)
    {
      errno = ENOENT;
      goto end;
    }

  -- poller_nfds;
  if (i < poller_nfds)
    {
      poller_fds[i] = poller_fds[poller_nfds];
      poller_data[i] = poller_data[poller_nfds];
    }

  /* Again, according to man pages, we need not to use free() explicitly.
     Quoting, 'If size is 0 and ptr is not a null pointer, the object
     pointed to is freed.'.  */

  poller_fds = realloc (poller_fds, sizeof (struct pollfd) * poller_nfds);
  poller_data = realloc (poller_data, sizeof (pollerdata_t) * poller_nfds);

end:
  return poller_nfds;
}

int
poller_poll (int timeout, pollercb_t callback)
{
  int i, res, have;

  res = poll (poller_fds, poller_nfds, timeout);
  if (res <= 0)
    return res;

  for (i = 0, have = 0; i < poller_nfds && have < res; i++)
    if (poller_fds[i].revents)
      {
        callback (poller_data[i], poller_fds[i].revents);
        have ++;
      }

  return res;
}

#endif /* not HAVE_EPOLL_WAIT */
